using System.Net;
using DotNext;
namespace RemoteUpdate;

static class Global
{
    /// <summary>
    /// FPGA 远程更新器地址
    /// </summary>
    public class RemoteUpdaterAddr
    {
        /// <summary> 
        /// 基地址
        /// </summary>
        public const UInt32 BaseAddr = 0x2000_0000;
        /// <summary> 
        /// 写比特流-读写地址——控制位 <br/>
        /// [31:24]: 保留 <br/>
        /// [23:17]: 保留 <br/>
        /// [16: 8]: 保留 <br/>
        /// [ 7: 0]: {-, -, -, -, -,{bitstream_wr_num}, flash_wr_en} <br/>
        /// </summary>
        public const UInt32 WriteCtrl = 0x2000_0000;
        /// <summary> 
        /// 写比特流-只写地址——FIFO入口
        /// [31:0]:  写比特流数据入口
        /// </summary>
        public const UInt32 WriteFIFO = 0x2000_0001;
        /// <summary> 
        /// 写比特流-只读地址——标志位 <br/>
        /// [31:24]: {-, -, -, -, -, -, -,      wr_fifo_full} <br/>
        /// [23:17]: {-, -, -, -, -, -, -,     wr_fifo_empty} <br/>
        /// [16: 8]: {-, -, -, -, -, -, -, bitstream_wr_done} <br/>
        /// [ 7: 0]: {-, -, -, -, -, -, -,     clear_bs_done} <br/>
        /// </summary>
        public const UInt32 WriteSign = 0x2000_0002;
        /// <summary> 
        /// 读比特流-读写地址——控制位 <br/>
        /// [31:24]: {-, -, -, -, -, -,{   bs_crc32_ok           }} <br/>
        /// [23:17]: {-, -, -, -, -, -, -,            crc_check_en} <br/>
        /// [16: 8]: {-, -, -, -, -, -, -,     bitstream_up2cpu_en} <br/>
        /// [ 7: 0]: {-, -, -, -, -,{bitstream_rd_num},flash_rd_en} <br/>
        /// </summary>
        public const UInt32 ReadCtrl = 0x2000_0003;
        /// <summary> 
        /// 读比特流-只读地址——FIFO出口
        /// [31:0]: 读比特流数据出口
        /// </summary>
        public const UInt32 ReadFIFO = 0x2000_0004;
        /// <summary> 
        /// 读比特流-只读地址——CRC校验值
        /// [31:0]: CRC校验值 bs_readback_crc
        /// </summary>
        public const UInt32 ReadCRC = 0x2000_0005;
        /// <summary> 
        /// 读比特流-只读地址——标志位 <br/>
        /// [31:24]: {-, -, -, -, -, -, -,         rd_fifo_afull} <br/>
        /// [23:17]: {-, -, -, -, -, -, -,         rd_fifo_empty} <br/>
        /// [16: 8]: {-, -, -, -, -, -, -,     bitstream_rd_done} <br/>
        /// [ 7: 0]: {-, -, -, -, -, -, -, bs_readback_crc_valid} <br/>
        /// </summary>
        public const UInt32 ReadSign = 0x2000_0006;
        /// <summary> 
        /// 单独擦除开关-读写地址——控制位
        /// [31:24]: {-, -, -, -, -, -, -,       -    }
        /// [23:17]: {-, -, -, -, -, -, -,       -    }
        /// [16: 8]: {-, -, -, -, -, -, -,       -    }
        /// [ 7: 0]: {-, -, -, -, -, -, -, clear_sw_en}
        /// </summary>
        public const UInt32 ClearCtrl = 0x2000_0007;
        /// <summary> 
        /// 单独擦除开关-只读地址——标志位 <br/>
        /// [31:24]: {-, -, -, -, -, -, -,  time_out_reg} <br/>
        /// [23:17]: {           flash_flag_status[15:8]} <br/>
        /// [16: 8]: {            flash_flag_status[7:0]} <br/>
        /// [ 7: 0]: {-, -, -, -, -, -, -, clear_sw_done} <br/>
        /// </summary>
        public const UInt32 ClearSign = 0x2000_0008;
        /// <summary> 
        /// 写开关-读写地址——标志位
        /// [31:24]: {-, -, -, -, -, -, -,       -         }
        /// [23:17]: {-, -, -, -, -, -, -,       -         }
        /// [16: 8]: {-, -, -, -, -, -, {      open_sw_num}}
        /// [ 7: 0]: {-, -, -, -, -, -, -, write_sw_code_en}
        /// </summary>
        public const UInt32 WriteSwitchSign = 0x2000_0009;
        /// <summary> 
        /// 写开关-只读地址——控制位
        /// [31:24]: {-, -, -, -, -, -, -,       -          }
        /// [23:17]: {-, -, -, -, -, -, -,       -          }
        /// [16: 8]: {-, -, -, -, -, -, -,         ipal_busy}
        /// [ 7: 0]: {-, -, -, -, -, -, -, open_sw_code_done}
        /// </summary>
        public const UInt32 WriteSwitchCtrl = 0x2000_0010;
        /// <summary> 
        /// 热启动开关-读写地址——控制位
        /// [31:24]: {-, -, -, -, -, -, -,       -       }
        /// [23:17]: {-, -, -, -, -, -, -,       -       }
        /// [16: 8]: {-, -, -, -, -, -,                  }
        /// [ 7: 0]: {-, -, -, -, -, -, -,    hotreset_en}
        /// </summary>
        public const UInt32 HotResetCtrl = 0x2000_0011;
    }

}

/// <summary>
/// FPGA远程更新
/// </summary>
public class RemoteUpdater
{
    private static readonly NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();

    readonly int timeout = 1000;

    int port { get; }
    string address { get; }
    private IPEndPoint remoteEP;

    /// <summary>
    /// 初始化 FPGA 远程更新器
    /// </summary>
    /// <param name="address">目标设备的 IP 地址</param>
    /// <param name="port">目标设备的端口号</param>
    public RemoteUpdater(string address, int port)
    {
        this.address = address;
        this.port = port;
        this.remoteEP = new IPEndPoint(IPAddress.Parse(address), port);
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <param name="bitstream">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> WriteBitstream(int bitstreamNum, byte[] bitstream)
    {
        if (bitstreamNum < 0 || bitstreamNum > 0b11)
            return new(new ArgumentException(
                "The number of bitstream can't be negative or over 3(0b11)", nameof(bitstreamNum)));

        if (bitstream.Length % (4 * 1024) != 0)
            return new(new ArgumentException(
                "Bitstream doesn't match 4KB align, which should be divided by 4 * 1024", nameof(bitstream)));

        // Clear Data
        await MsgBus.UDPServer.ClearUDPData(this.address);

        logger.Trace($"Clear up udp server {this.address} receive data");


        {
            var ret = await UDPClientPool.WriteAddr(
                    this.remoteEP, Global.RemoteUpdaterAddr.WriteCtrl, (UInt32)((bitstreamNum << 1) | 0b1), timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Enable write flash failed"));
        }
        // {
        //     var ret = await UDPClientPool.ReadAddrWithWait(
        //             this.remoteEP, Global.RemoteUpdaterAddr.ClearSign, 0x00_01, 0x00_01, 60 * 1000);
        //     if (!ret.IsSuccessful) return new(ret.Error);
        //     if (!ret.Value) return new(new Exception("Clear switch failed even over time"));
        // }
        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                    this.remoteEP, Global.RemoteUpdaterAddr.WriteSign, 0x00_01, 0x00_01, 60 * 1000);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Clear bitstream failed even over time"));
        }
        logger.Trace($"Device {this.address} remote update init finished");

        {
            var ret = await UDPClientPool.WriteAddr(this.remoteEP, Global.RemoteUpdaterAddr.WriteFIFO, bitstream);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Send bitstream failed"));
        }
        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                    this.remoteEP, Global.RemoteUpdaterAddr.WriteSign, 0x00_10, 0x00_10, timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Write bitstream failed even over time"));
        }
        logger.Trace($"Device {this.address} write bitstream finished");

        return true;
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <param name="checkSum">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> CheckBitstreamCRC(int bitstreamNum, UInt32 checkSum)
    {
        // Clear Data
        await MsgBus.UDPServer.ClearUDPData(this.address);

        logger.Trace($"Clear up udp server {this.address} receive data");


        {
            var ret = await UDPClientPool.WriteAddr(
                this.remoteEP, Global.RemoteUpdaterAddr.ReadCtrl,
                (UInt32)((bitstreamNum << 1) | 0b1), timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Enable read failed"));
        }
        {
            var ret = await UDPClientPool.ReadAddrWithWait(
                    this.remoteEP, Global.RemoteUpdaterAddr.ReadSign, 0x00_11, 0x00_11, 60 * 1000);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Read bitstream failed even over time"));
        }
        logger.Trace($"Device {this.address} remote update read bitstream finished");

        {
            var ret = await UDPClientPool.ReadAddr(
                    this.remoteEP, Global.RemoteUpdaterAddr.ReadSign, timeout);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (ret.Value.Options.Data is null) return new(new Exception("Receive null when read sign"));

            // Check CRC
            var crcCode = Common.Number.BytesToUInt32(ret.Value.Options.Data);
            if (!crcCode.IsSuccessful) return new(crcCode.Error);

            var retCRC = crcCode.Value == checkSum;
            logger.Trace($"Device {this.address} remote update check crc finished");

            return retCRC;
        }
    }

    /// <summary>
    /// [TODO:description]
    /// </summary>
    /// <param name="bitstreamNum">[TODO:parameter]</param>
    /// <param name="bitstream">[TODO:parameter]</param>
    /// <returns>[TODO:return]</returns>
    public async ValueTask<Result<bool>> UpdateBitstream(int bitstreamNum, byte[] bitstream)
    {
        {
            var ret = await WriteBitstream(bitstreamNum, bitstream);
            if (!ret.IsSuccessful) return new(ret.Error);
            if (!ret.Value) return new(new Exception("Write bitstream failed"));
        }
        {
            var crc = new System.IO.Hashing.Crc32();
            crc.Append(bitstream);
            var ret = await CheckBitstreamCRC(bitstreamNum, crc.GetCurrentHashAsUInt32());
            if (!ret.IsSuccessful) return new(ret.Error);

            return ret.Value;
        }
    }

}
